# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flax layer for AQT injection."""

import copy
import enum
import functools
from typing import Iterable
from typing import Optional
from . import q_dot_general
from . import calibration
from . import config
from . import int_numerics
from . import no_numerics
import flax
import flax.linen as nn
import jax
from jax._src.numpy import lax_numpy
import jax.numpy as jnp


class QuantMode(enum.Enum):
    TRAIN = 1
    CONVERT = 2
    SERVE = 3


class Freezer(nn.Module):
    """Identity function that can freeze its input.

    On default it is an identity function that saves the input in a variable.
    In 'use_frozen=True' mode, ignores the input and returns the frozen value. It
    is usefult to implement 'constant folding' and put quantized weights and
    scales in the checkpoint for serving.
    """

    quant_collection: str
    quant_mode: QuantMode
    q_shape: Iterable[int]
    q_init: nn.initializers.Initializer
    s_shape: Iterable[int]
    s_init: nn.initializers.Initializer

    @nn.compact
    def __call__(
            self, inputs: Optional[q_dot_general.QTensor]
    ) -> Optional[q_dot_general.QTensor]:
        collection = self.quant_collection
        if inputs is None:  # getter mode
            if self.quant_mode == QuantMode.TRAIN:
                return inputs
            elif self.quant_mode == QuantMode.CONVERT:
                return inputs
            elif self.quant_mode == QuantMode.SERVE:
                # We could have created one self.variable whose value is a QTensor,
                # but this would complicate the init function, which could potentially
                # be used by adding metadata such as sharding axises, etc.
                qvalue = self.variable(collection, 'value', self.q_init, self.q_shape)
                scale = self.variable(collection, 'scale', self.s_init, self.s_shape)
                return q_dot_general.QTensor(qvalue.value, scale.value)
            else:
                assert False, 'Unknown quant mode.'
        else:  # setter mode
            if self.quant_mode == QuantMode.TRAIN:
                pass
            elif self.quant_mode == QuantMode.CONVERT:
                qvalue = self.variable(collection, 'value', self.q_init, self.q_shape)
                scale = self.variable(collection, 'scale', self.s_init, self.s_shape)
                qvalue.value = inputs.qvalue
                scale.value = inputs.qvalue_scale_t
            elif self.quant_mode == QuantMode.SERVE:
                # TODO(lew): Optionally compare stored and served value.
                pass
            else:
                assert False, 'Unknown quant mode.'
            return None


class QDotGeneral(nn.Module):
    """A layer that can be injected into flax.nn.Dense, etc."""

    cfg: Optional[config.DotGeneral] = None
    prng_name: Optional[str] = 'params'

    # TODO(lew): split out separate class for each side.
    lhs_quant_mode: QuantMode = QuantMode.TRAIN
    lhs_init: nn.initializers.Initializer = jnp.zeros
    lhs_scale_init: nn.initializers.Initializer = jnp.zeros
    lhs_var_name: str = 'qlhs'

    rhs_quant_mode: QuantMode = QuantMode.TRAIN
    rhs_init: nn.initializers.Initializer = jnp.zeros
    rhs_scale_init: nn.initializers.Initializer = jnp.zeros
    rhs_var_name: str = 'qrhs'

    # If you want use 'params' make sure that there is another mechanism to hide
    # these variables from the optimizer.
    quant_collection: str = 'aqt'

    def make_aqt_dg(
            self,
            lhs_shape,
            rhs_shape,
            dimension_numbers: tuple[Iterable[int], Iterable[int]],
    ):
        lhs_scale_shape = list(lhs_shape)
        rhs_scale_shape = list(rhs_shape)
        (contr, _) = dimension_numbers
        for li, ri in zip(*contr):
            lhs_scale_shape[li] = 1
            rhs_scale_shape[ri] = 1
        lhs_scale = q_dot_general._lhs_scale_transpose(  # pylint: disable=protected-access
            jnp.zeros(lhs_scale_shape), dimension_numbers, lhs_shape, rhs_shape
        )
        assert lhs_scale is not None
        lhs_scale_shape = lhs_scale.shape
        rhs_scale = q_dot_general._rhs_scale_transpose(  # pylint: disable=protected-access
            jnp.zeros(rhs_scale_shape), dimension_numbers, lhs_shape, rhs_shape
        )
        assert rhs_scale is not None
        rhs_scale_shape = rhs_scale.shape

        cfg = copy.deepcopy(self.cfg)
        if cfg is not None:
            rhs_qm = self.rhs_quant_mode
            lhs_qm = self.lhs_quant_mode

            msg = 'The only function that is setting preprocess can be QQuantized.'
            assert cfg.fwd.rhs.preprocess is None, msg
            assert cfg.fwd.lhs.preprocess is None, msg
            cfg.fwd.lhs.preprocess = Freezer(
                name=self.lhs_var_name,
                quant_mode=lhs_qm,
                q_shape=lhs_shape,
                q_init=self.lhs_init,
                s_shape=lhs_scale_shape,
                s_init=self.lhs_scale_init,
                quant_collection=self.quant_collection,
            )
            cfg.fwd.rhs.preprocess = Freezer(
                name=self.rhs_var_name,
                quant_mode=rhs_qm,
                q_shape=rhs_shape,
                q_init=self.rhs_init,
                s_shape=rhs_scale_shape,
                s_init=self.rhs_scale_init,
                quant_collection=self.quant_collection,
            )
        key = self.make_rng(self.prng_name) if self.prng_name is not None else None
        context = q_dot_general.Context(key=key, train_step=None)
        aqt_dg = q_dot_general.make_dot_general(cfg)
        aqt_dg = functools.partial(aqt_dg, context=context)
        return aqt_dg

    @nn.compact
    def __call__(
            self,
            lhs,
            rhs,
            dimension_numbers,
            precision,
            preferred_element_type=None,
    ):
        aqt_dg = self.make_aqt_dg(lhs.shape, rhs.shape, dimension_numbers)
        return aqt_dg(
            lhs,
            rhs,
            dimension_numbers,
            precision,
            preferred_element_type=preferred_element_type,
        )


class QEinsum(flax.struct.PyTreeNode):
    """Quantized Einsum class for model injection."""

    cfg: Optional[config.DotGeneral] = None
    prng_name: Optional[str] = 'params'

    # TODO(lew): split out separate class for each side.
    lhs_quant_mode: QuantMode = QuantMode.TRAIN
    lhs_init: nn.initializers.Initializer = jnp.zeros
    lhs_scale_init: nn.initializers.Initializer = jnp.zeros
    lhs_var_name: str = 'qlhs'

    rhs_quant_mode: QuantMode = QuantMode.TRAIN
    rhs_init: nn.initializers.Initializer = jnp.zeros
    rhs_scale_init: nn.initializers.Initializer = jnp.zeros
    rhs_var_name: str = 'qrhs'

    # If you want use 'params' make sure that there is another mechanism to hide
    # these variables from the optimizer.
    quant_collection: str = 'aqt'

    def __call__(self, eqn, lhs_g, rhs_g):
        def einsum(lhs_l, rhs_l, dg=jax.lax.dot_general):
            operands, contractions = lax_numpy._default_poly_einsum_handler(  # pylint: disable=protected-access
                eqn, lhs_l, rhs_l, einsum_call=True, use_blas=True, optimize='optimal'
            )
            contractions = tuple((a, frozenset(b), c) for a, b, c, *_ in contractions)
            return jax.named_call(lax_numpy._einsum, name=eqn)(  # pylint: disable=protected-access
                operands,
                contractions,
                precision=None,
                preferred_element_type=None,
                _dot_general=dg,
            )

        # yes_swap = whether einsum swaps [lhs,rhs] when passing them to dot_general
        a = jax.make_jaxpr(einsum)(lhs_g, rhs_g)
        [lhs_g_id, rhs_g_id] = a.eqns[0].invars
        [lhs_l_id, rhs_l_id] = a.jaxpr.invars
        not_swap = lhs_g_id == lhs_l_id and rhs_g_id == rhs_l_id
        yes_swap = lhs_g_id == rhs_l_id and rhs_g_id == lhs_l_id
        assert not_swap != yes_swap

        cfg = copy.deepcopy(self.cfg)
        prng_name = self.prng_name

        lhs_quant_mode = self.lhs_quant_mode
        lhs_init = self.lhs_init
        lhs_scale_init = self.lhs_scale_init
        lhs_var_name = self.lhs_var_name

        rhs_quant_mode = self.rhs_quant_mode
        rhs_init = self.rhs_init
        rhs_scale_init = self.rhs_scale_init
        rhs_var_name = self.rhs_var_name

        quant_collection = self.quant_collection

        if yes_swap:
            if cfg is not None:
                cfg.fwd.lhs, cfg.fwd.rhs = cfg.fwd.rhs, cfg.fwd.lhs
                cfg.dlhs, cfg.drhs = cfg.drhs, cfg.dlhs
            lhs_quant_mode, rhs_quant_mode = rhs_quant_mode, lhs_quant_mode
            lhs_init, rhs_init = rhs_init, lhs_init
            lhs_scale_init, rhs_scale_init = rhs_scale_init, lhs_scale_init
            lhs_var_name, rhs_var_name = rhs_var_name, lhs_var_name

        aqt_dg = QDotGeneral(
            cfg=cfg,
            prng_name=prng_name,
            lhs_quant_mode=lhs_quant_mode,
            lhs_init=lhs_init,
            lhs_scale_init=lhs_scale_init,
            lhs_var_name=lhs_var_name,
            rhs_quant_mode=rhs_quant_mode,
            rhs_init=rhs_init,
            rhs_scale_init=rhs_scale_init,
            rhs_var_name=rhs_var_name,
            quant_collection=quant_collection,
        )
        return einsum(lhs_g, rhs_g, aqt_dg)


def config_v4(
        *,
        fwd_bits: Optional[int] = 8,
        dlhs_bits: Optional[int] = 8,
        drhs_bits: Optional[int] = None,
        # The dummy static bound flag is for performance benchmarking.
        use_dummy_static_bound: bool = False,
        rng_type: str = 'jax.uniform',  # 'custom-1'
        dlhs_local_aqt: Optional[config.LocalQ] = None,
        drhs_local_aqt: Optional[config.LocalQ] = None,
        fwd_accumulator_dtype: ... = jnp.int32,
        dlhs_accumulator_dtype: ... = jnp.int32,
        drhs_accumulator_dtype: ... = None,
) -> config.DotGeneral:
    """
    The config_v4 function is a helper function that creates a DotGeneral config
    object. It takes in the following arguments:
    - fwd_bits: The number of bits to use for forward pass quantization. If None, no quantization will be used.
    Defaults to 8 bits.
    - dlhs_bits: The number of bits to use for left hand side gradient quantization (i.e., the weights). If None,
     no quantization will be used. Defaults to 8 bits..
    - drhs_bits: The number of bits to use for right hand side gradient quanitzation

    :param *: Indicate that the function accepts a variable number of arguments
    :param fwd_bits: Optional[int]: Set the number of bits for the forward pass
    :param dlhs_bits: Optional[int]: Set the number of bits used for quantization
    :param drhs_bits: Optional[int]: Set the number of bits for the right hand side
    :param use_dummy_static_bound: bool: Set the static bound to 1
    :param rng_type: str: Set the type of random number generator
    :param dlhs_local_aqt: Optional[config.LocalQ]: Set the local quantization parameters for the lhs gradient
    :param drhs_local_aqt: Optional[config.LocalQ]: Set the local quantization parameters for the drhs tensor
    :param fwd_accumulator_dtype: ...: Set the accumulator dtype for forward pass
    :param dlhs_accumulator_dtype: ...: Determine the dtype of the accumulator in
    :param drhs_accumulator_dtype: ...: Determine the dtype of the accumulator used in q_dot_general
    :param : Determine the number of bits used for quantization
    :return: A config
    """

    def tensor_config(bits: Optional[int]) -> config.Tensor:
        assert bits is None or bits >= 2, 'Need at least 2 bits.'
        if bits is None:
            numerics = no_numerics.NoNumerics()
        else:
            numerics = int_numerics.IntNumerics(
                bits=bits,
                preserve_zero=True,
                preserve_max_val=False,
                clip=True,
                round=True,
                noise_fn=None,
                clip_gradient=False,  # Can be False when using abs-max scaling.
                dtype=jnp.int8 if 2 <= bits <= 8 else None,
            )

        return config.Tensor(
            numerics=numerics,
            calib_shared_axes=None,
            scale_stop_grad=True,
            calibration=calibration.AbsMaxCalibration(),
            po2_scale=False,
            use_fake_quant=False,
            # dtype_x=dtype,
            use_fwd_quant=None,
            preprocess=None,
        )

    def dg_raw_config(lhs_bits, rhs_bits, local_aqt=None) -> config.DotGeneralRaw:
        lhs_cfg = tensor_config(lhs_bits)
        rhs_cfg = tensor_config(rhs_bits)
        if (
                True  # Just to format lines below
                and lhs_bits is not None
                and rhs_bits is not None
                and lhs_bits <= 8
                and rhs_bits <= 8
        ):
            dg_accumulator_dtype = jnp.int32
        else:
            # None determines the dtype on the fly in q_dot_general
            dg_accumulator_dtype = None

        return config.DotGeneralRaw(
            lhs=lhs_cfg,
            rhs=rhs_cfg,
            dg_accumulator_dtype=dg_accumulator_dtype,
            local_aqt=local_aqt,
        )

    cfg = config.DotGeneral(
        fwd=dg_raw_config(fwd_bits, fwd_bits),
        dlhs=dg_raw_config(dlhs_bits, dlhs_bits, local_aqt=dlhs_local_aqt),
        drhs=dg_raw_config(drhs_bits, drhs_bits, local_aqt=drhs_local_aqt),
    )

    cfg.dlhs.rhs.use_fwd_quant = False
    cfg.drhs.rhs.use_fwd_quant = False

    # Typically we have (but I don't know if it is guraranteed):
    # - vjp_lhs_stochastic_rounding is referring to the gradient and
    # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
    config.set_stochastic_rounding(
        cfg,
        vjp_lhs_stochastic_rounding=True,
        vjp_rhs_stochastic_rounding=False,
        implementation=rng_type,
    )

    if use_dummy_static_bound:
        config.set_static_bound(cfg, 1.0)

    config.set_accumulator_dtype(
        cfg,
        fwd_dtype=fwd_accumulator_dtype,
        dlhs_dtype=dlhs_accumulator_dtype,
        drhs_dtype=drhs_accumulator_dtype,
    )

    return cfg



import curses
import json
import subprocess
import time

import IPython.display
import jax
import threading
import os

try:
    import posix
except ModuleNotFoundError:
    posix = None


def is_notebook():
    """Returns True if the code is being run in a notebook, False otherwise."""
    return os.environ.get("IPYTHON") is not None


# Edited version of Jax-SMI from https://github.com/ayaka14732/jax-smi/
def run(note_book=None, interval: float = 1, dir_prefix: str = '/dev/shm', dpr=True):
    """
    The run function is a simple wrapper around the go tool pprof command.
    It runs the command every interval seconds and prints out its output to stdout.
    If you are running this in a Jupyter notebook, it will print to an IPython display object instead of stdout.


    :param note_book: Determine whether the program is running in a notebook or not
    :param interval: float: Specify the interval between each refresh
    :param dir_prefix: str: Specify the directory where the memory
    :param dpr: Determine whether to display the output in a notebook or not
    :return: The output of the pprof command

    """
    if note_book is None:
        note_book = is_notebook()
    std = curses.initscr() if not note_book else None
    try:
        while True:
            if not note_book and dpr:
                std.clear()
            output = subprocess.run(
                args=['go', 'tool', 'pprof', '-tags', f'{dir_prefix}/memory.prof'],
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
            ).stdout.decode('utf-8')
            if not note_book and dpr:
                std.addstr(output)
                std.refresh()
            if note_book and dpr:
                IPython.display.clear_output(True)
                print(output)

            with open(f'{dir_prefix}/memory.json', 'w') as fin:
                json.dump({
                    'log': output
                }, fin)
            time.sleep(interval)
    except KeyboardInterrupt:
        curses.endwin()


def get_mem(dir_prefix: str = '/dev/shm') -> str:
    """
    The get_mem function is a wrapper around the go tool pprof command.
    It takes in an optional argument, dir_prefix, which defaults to /dev/shm.
    The function then runs the go tool pprof command with arguments -tags and {dir_prefix}/memory.prof.
    The output of this command is captured and returned as a string.

    :param dir_prefix: str: Specify the directory prefix for
    :return: A string that contains the memory profile

    """
    return subprocess.run(
        args=['go', 'tool', 'pprof', '-tags', f'{dir_prefix}/memory.prof'],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    ).stdout.decode('utf-8')


def initialise_tracking(interval: float = 1., dir_prefix: str = '/dev/shm') -> None:
    """
    The initialise_tracking function starts a daemon thread that periodically saves the memory profile to disk.
    The outer function starts the daemon thread and returns a context manager that stops it when
    the context exits.  The inner function uses posix.rename() to atomically replace an existing file,
    so we can be sure that any given memory profile was taken at some point during the lifetime of its
    context.

    :param interval: float: Set the time interval between each memory profile
    :param dir_prefix: str: Specify the directory where the memory profile will be saved
    :return: A thread object

    """

    def inner():
        """
        The inner function is a daemon thread that periodically saves the memory profile to disk.
        The outer function starts the daemon thread and returns a context manager that stops it when
        the context exits.  The inner function uses posix.rename() to atomically replace an existing file,
        so we can be sure that any given memory profile was taken at some point during the lifetime of its
        context.

        :return: A thread object

        """
        while True:
            jax.profiler.save_device_memory_profile(f'{dir_prefix}/memory.prof.new')
            if posix is not None:
                os.rename(f'{dir_prefix}/memory.prof.new', f'{dir_prefix}/memory.prof')
            else:
                posix.rename(f'{dir_prefix}/memory.prof.new', f'{dir_prefix}/memory.prof')
            time.sleep(interval)

    thread = threading.Thread(target=inner, daemon=True)
    thread.start()


def threaded_log(interval: float = 1., dir_prefix: str = '/dev/shm', save_mem_json: bool = False) -> threading.Thread:
    """
    The threaded_log function is a wrapper around the get_mem function.
    It allows you to monitor your memory usage in real time, and optionally save it to a JSON file.
    The threaded_log function returns a threading.Thread object that can be started with .start() and stopped with .join().


    :param interval: float: Set the time interval between each memory log
    :param dir_prefix: str: Specify the directory to save the memory
    :param save_mem_json: bool: Save the memory information to a json file
    :return: A threading

    """
    note_book = is_notebook()

    def show_():

        std = curses.initscr() if not note_book else None
        try:
            while True:
                mem_info = get_mem()
                if not note_book:
                    std.clear()
                    std.addstr(mem_info)
                    std.refresh()
                if note_book:
                    IPython.display.clear_output(True)
                    print(mem_info)
                if save_mem_json:
                    with open(f'{dir_prefix}/memory.json', 'w') as fin:
                        json.dump({
                            'log': mem_info
                        }, fin)
                time.sleep(interval)
        except KeyboardInterrupt:
            curses.endwin()

    thread = threading.Thread(
        target=show_
    )
    return thread



import os

from ml_collections import ConfigDict
import jax
import flax
from flax.serialization import (
    from_bytes, to_bytes, to_state_dict, from_state_dict
)
from flax.traverse_util import flatten_dict, unflatten_dict, empty_node
import msgpack
from jax import numpy as jnp


def get_dtype(tensor, dtype):
    if dtype is None or dtype == '':
        return tensor
    if isinstance(dtype, str):
        dtype = {
            'bf16': jnp.bfloat16,
            'bfloat16': jnp.bfloat16,
            'fp16': jnp.float16,
            'float16': jnp.float16,
            'fp32': jnp.float32,
            'float32': jnp.float32,
            'fp64': jnp.float64,
            'float64': jnp.float64,
        }[dtype]
    float_dtypes = (jnp.bfloat16, jnp.float16, jnp.float32, jnp.float64)
    if getattr(tensor, 'dtype', None) in float_dtypes:
        tensor = tensor.astype(dtype)
    return tensor


class StreamingCheckpointer(object):
    """ Custom msgpack checkpointer that saves large train states by serializing
        and saving tensors one by one in a streaming fashion. Avoids running
        out of memory or local TPU disk with default flax checkpointer.
    """

    @staticmethod
    def get_default_config(updates=None):
        config = ConfigDict()
        config.float_dtype = 'bf16'
        config.save_optimizer_state = False

        if updates is not None:
            config.update(ConfigDict(updates).copy_and_resolve_references())
        return config

    def __init__(self, config, checkpoint_dir, enable=True):
        self.config = self.get_default_config(config)
        self.checkpoint_dir = checkpoint_dir
        self.enable = enable

    def save_checkpoint(self, train_state, filename, gather_fns=None):
        if self.enable:
            path = os.path.join(self.checkpoint_dir, filename)
        else:
            path = '/dev/null'
        self.save_train_state_to_file(
            train_state, path, gather_fns, self.config.float_dtype
        )

    @staticmethod
    def save_train_state_to_file(train_state, path, gather_fns=None, float_dtype=None):
        train_state = to_state_dict(train_state)
        packer = msgpack.Packer()
        flattend_train_state = flatten_dict(train_state)
        if gather_fns is not None:
            gather_fns = flatten_dict(to_state_dict(gather_fns))

        with open(path, "wb") as fout:
            for key, value in flattend_train_state.items():
                if gather_fns is not None:
                    value = gather_fns[key](value)
                value = get_dtype(value, float_dtype)
                fout.write(packer.pack((key, to_bytes(value))))

    def save_pickle(self, obj, filename):
        """
        The save_pickle function saves a Python object to disk using the pickle module.

        :param self: Represent the instance of the class
        :param obj: Pass the object that is to be pickled
        :param filename: Specify the name of the file to be saved
        :return: A pickle object
        
        """
        import pickle
        def save_pickle(obj, path):
            with open(path, 'wb') as fout:
                pickle.dump(obj, fout)

        if self.enable:
            path = os.path.join(self.checkpoint_dir, filename)
        else:
            path = '/dev/null'
        save_pickle(obj, path)

    def save_all(self, train_state, gather_fns, metadata=None, dataset=None, milestone=False):
        """
        The save_all function saves the following:
            - metadata.pkl (a pickle file containing a dictionary of metadata)
            - dataset.pkl (a pickle file containing the training data)
            - streaming_params_{step}.pkl or streaming_train_state_{step}.pkl
                (depending on whether we want to save optimizer state or not,
                this is a checkpoint that will not be overwritten by future checkpoints)

        :param self: Access the attributes and methods of the class
        :param train_state: Save the current state of the model
        :param gather_fns: Gather the state of the optimizer
        :param metadata: Save the metadata of the training
        :param dataset: Save the dataset to disk
        :param milestone: Determine whether the checkpoint is a milestone or not
        :return: Nothing
        
        """
        step = int(jax.device_get(train_state.step))
        if self.config.save_optimizer_state:
            checkpoint_state = train_state
            checkpoint_name = 'streaming_train_state'
            checkpoint_gather_fns = gather_fns
        else:
            checkpoint_state = train_state.params['params']
            checkpoint_name = 'streaming_params'
            checkpoint_gather_fns = gather_fns.params['params']

        if milestone:
            # Save a milestone checkpoint that will not be overwritten
            self.save_pickle(metadata, f'metadata_{step}.pkl')
            self.save_pickle(dataset, f'dataset_{step}.pkl')
            self.save_checkpoint(
                checkpoint_state, f'{checkpoint_name}_{step}', checkpoint_gather_fns
            )
        else:
            # Save a normal checkpoint that can be overwritten
            self.save_pickle(metadata, 'metadata.pkl')
            self.save_pickle(dataset, 'dataset.pkl')
            self.save_checkpoint(
                checkpoint_state, f'{checkpoint_name}', checkpoint_gather_fns
            )

    @staticmethod
    def load_checkpoint(path, target=None, shard_fns=None, remove_dict_prefix=None):
        """
        The load_checkpoint function is used to load a checkpoint from disk.

        :param path: Specify the path to the checkpoint file
        :param target: Specify the model to load the checkpoint into
        :param shard_fns: Specify a function that will be applied to each tensor in the checkpoint
        :param remove_dict_prefix: Remove the prefix of a dictionary
        :return: A dictionary of the form {key: value}, where key is a tuple and value is a tensor
        
        """
        if shard_fns is not None:
            shard_fns = flatten_dict(
                to_state_dict(shard_fns)
            )
        if remove_dict_prefix is not None:
            remove_dict_prefix = tuple(remove_dict_prefix)
        flattend_train_state = {}
        with open(path, 'rb') as fin:
            # 83886080 bytes = 80 MB, which is 16 blocks on GCS
            unpacker = msgpack.Unpacker(fin, read_size=83886080, max_buffer_size=0)
            for key, value in unpacker:
                key = tuple(key)
                if remove_dict_prefix is not None:
                    if key[:len(remove_dict_prefix)] == remove_dict_prefix:
                        key = key[len(remove_dict_prefix):]
                    else:
                        continue

                tensor = from_bytes(None, value)
                if shard_fns is not None:
                    tensor = shard_fns[key](tensor)
                flattend_train_state[key] = tensor

        if target is not None:
            flattened_target = flatten_dict(
                to_state_dict(target), keep_empty_nodes=True
            )
            for key, value in flattened_target.items():
                if key not in flattend_train_state and value == empty_node:
                    flattend_train_state[key] = value

        train_state = unflatten_dict(flattend_train_state)
        if target is None:
            return train_state

        return from_state_dict(target, train_state)

    @staticmethod
    def load_flax_checkpoint(path, target=None, shard_fns=None):
        """ Load a standard flax checkpoint that's not saved with the
            msgpack streaming format.
        """
        with open(path, "rb") as fin:
            encoded_bytes = fin.read()

        state_dict = flax.serialization.msgpack_restore(encoded_bytes)
        if shard_fns is not None:
            shard_fns = to_state_dict(shard_fns)
            state_dict = jax.tree_util.tree_map(lambda fn, x: fn(x), shard_fns, state_dict)

        if target is None:
            return state_dict
        return from_state_dict(target, state_dict)

    @classmethod
    def load_trainstate_checkpoint(cls, load_from, trainstate_target=None,
                                   trainstate_shard_fns=None,
                                   disallow_trainstate=False):
        """
        The load_trainstate_checkpoint function is used to load a checkpoint from disk.

        :param cls: Call the load_checkpoint function
        :param load_from: Specify where to load the model from
        :param trainstate_target: Specify the target for the train state
        :param trainstate_shard_fns: Specify the sharding function
        :param disallow_trainstate: Prevent loading the entire trainstate
        :return: A tuple of two objects, the train_state and restored_params
        
        """
        if trainstate_target is not None:
            params_target = trainstate_target.params['params']
        else:
            params_target = None

        if trainstate_shard_fns is not None:
            params_shard_fns = trainstate_shard_fns.params['params']
        else:
            params_shard_fns = None

        load_type, load_path = load_from.split('::', 1)
        if disallow_trainstate:
            assert load_type != 'trainstate', 'Loading full trainstate is not allowed!'
        train_state = None
        restored_params = None
        if load_type == 'trainstate':
            # Load the entire train state in the streaming format
            train_state = cls.load_checkpoint(
                path=load_path,
                target=trainstate_target,
                shard_fns=trainstate_shard_fns,
            )
        elif load_type == 'trainstate_params':
            # Load the params part of the train state in the streaming format
            restored_params = cls.load_checkpoint(
                path=load_path,
                target=params_target,
                shard_fns=params_shard_fns,
                remove_dict_prefix=('params', 'params'),
            )
            restored_params = flax.core.frozen_dict.freeze(
                {'params': restored_params}
            )
        elif load_type == 'params':
            # Load the params in the streaming format
            restored_params = cls.load_checkpoint(
                path=load_path,
                target=params_target,
                shard_fns=params_shard_fns,
            )
            restored_params = flax.core.frozen_dict.freeze(
                {'params': restored_params}
            )
        elif load_type == 'flax_params':
            # Load the params in the standard flax format (non-streaming)
            # This requires the entire params to fit in memory
            restored_params = cls.load_flax_checkpoint(
                path=load_path,
                target=params_target,
                shard_fns=params_shard_fns
            )
            restored_params = flax.core.frozen_dict.freeze(
                {'params': restored_params}
            )
        else:
            raise ValueError(f'Invalid load_from type: {load_type}')

        return train_state, restored_params



import jax
import jax.numpy as jnp
import re

from jax.experimental.mesh_utils import create_device_mesh
from jax.experimental.pjit import pjit, with_sharding_constraint as _with_sharding_constraint
import numpy as np
from jax.sharding import PartitionSpec as PS
from jax.experimental import mesh_utils
from jax.interpreters import pxla
import flax
from jax.sharding import Mesh
from typing import Sequence


def make_shard_and_gather_fns(partition_specs, dtype_specs=None):
    """
    The make_shard_and_gather_fns function takes in a partition_specs and dtype_specs,
    and returns two functions: shard_fns and gather_fns. The shard function is used to
    shard the input tensor into the specified partitions. The gather function is used to
    gather all the shards back together into one tensor.

    :param partition_specs: Specify the sharding of the input tensor
    :param dtype_specs: Specify the dtype of the tensor
    :return: A tuple of functions
    
    """
    float_dtypes = (jnp.bfloat16, jnp.float16, jnp.float32, jnp.float64)

    def make_to_dtype_fn(dtype_spec):
        def to_dtype(tensor):
            if dtype_specs in float_dtypes and getattr(tensor, 'dtype', None) in float_dtypes:
                # force np array to jax numpy array
                return jnp.asarray(tensor).astype(dtype_specs)
            elif hasattr(dtype_spec, 'dtype') and hasattr(tensor, 'dtype'):
                return jnp.asarray(tensor).astype(dtype_spec.dtype)
            return jnp.asarray(tensor)

        return to_dtype

    def make_shard_fn(partition_spec, dtype_spec=None):
        jax_shard_function = pjit(
            make_to_dtype_fn(dtype_spec),
            in_shardings=None,
            out_shardings=partition_spec
        )

        def shard_fn(tensor):
            return jax_shard_function(tensor).block_until_ready()

        return shard_fn

    def make_gather_fn(partition_spec, dtype_spec=None):
        jax_gather_fn = pjit(
            make_to_dtype_fn(dtype_spec),
            in_shardings=partition_spec,
            out_shardings=None
        )

        def gather_fn(tensor):
            return jax.device_get(jax_gather_fn(tensor))

        return gather_fn

    if dtype_specs is None or dtype_specs in float_dtypes:
        shard_fns = jax.tree_util.tree_map(make_shard_fn, partition_specs)
        gather_fns = jax.tree_util.tree_map(make_gather_fn, partition_specs)
    else:
        shard_fns = jax.tree_util.tree_map(
            make_shard_fn, partition_specs, dtype_specs
        )
        gather_fns = jax.tree_util.tree_map(
            make_gather_fn, partition_specs, dtype_specs
        )
    return shard_fns, gather_fns


def get_jax_mesh(axis_dims, names):
    """
    The get_jax_mesh function takes a string of the form:
        &lt;axis_dims&gt;
    where axis_dims is a comma-separated list of dimensions, each dimension being either:
        &lt;name&gt;:&lt;dim&gt;  or  &lt;dim&gt;
    If there are no names, then the default names 'x', 'y', and 'z' will be used. If there are fewer than three dimensions, then the remaining dimensions will be set to 1. For example:

    :param axis_dims: Specify the dimensions of the mesh
    :param names: Specify the names of the dimensions in
    :return: A mesh object
    
    """
    if axis_dims.startswith('!'):
        mesh_axis_splitting = True
        axis_dims = axis_dims[1:]
    else:
        mesh_axis_splitting = False

    if ':' in axis_dims:
        dims = []
        dim_names = []
        for axis in axis_dims.split(','):
            name, dim = axis.split(':')
            assert name in names
            dims.append(int(dim))
            dim_names.append(name)
        assert (set(dim_names) == set(names))
    else:
        dims = [int(x) for x in axis_dims.split(',')]
        dim_names = names
    assert len(dims) == len(names)
    mesh_shape = np.arange(jax.device_count()).reshape(dims).shape
    if mesh_axis_splitting:
        physical_mesh = np.array(jax.devices()).reshape(mesh_shape)
    else:
        physical_mesh = mesh_utils.create_device_mesh(mesh_shape)
    return Mesh(physical_mesh, dim_names)


def names_in_current_mesh(*names):
    """
    The names_in_current_mesh function is used to check if a set of names are in the current mesh.

    :param *names: Pass in a list of names to the function
    :return: A boolean indicating whether
    
    """
    mesh_axis_names = pxla.thread_resources.env.physical_mesh.axis_names
    return set(names) <= set(mesh_axis_names)


def get_names_from_partition_spec(partition_specs):
    """
    The get_names_from_partition_spec function takes a partition_specs argument, which is either a dictionary or list.
    If it's a dictionary, the function converts it to a list of values. Then for each item in the partition_specs list:
        If the item is None, continue (do nothing) and move on to next iteration of loop.
        If the item is an instance of str (i.e., if it's just one string), add that string to names set and move on to next iteration of loop.
        Otherwise (if not None or str), call get_names_from_partition_spec recurs

    :param partition_specs: Specify the partitioning of the data
    :return: A list of names
    
    """
    names = set()
    if isinstance(partition_specs, dict):
        partition_specs = partition_specs.values()
    for item in partition_specs:
        if item is None:
            continue
        elif isinstance(item, str):
            names.add(item)
        else:
            names.update(get_names_from_partition_spec(item))

    return list(names)


def with_sharding_constraint(x, partition_specs):
    """ A smarter version of with_sharding_constraint that only applies the
        constraint if the current mesh contains the axes in the partition specs.
    """
    axis_names = get_names_from_partition_spec(partition_specs)
    if names_in_current_mesh(*axis_names):
        x = _with_sharding_constraint(x, partition_specs)
    return x


def wrap_function_with_rng(rng):
    """ To be used as decorator, automatically bookkeep a RNG for the wrapped function. """

    def wrap_function(function):
        def wrapped(*args, **kwargs):
            nonlocal rng
            rng, split_rng = jax.random.split(rng)
            return function(split_rng, *args, **kwargs)

        return wrapped

    return wrap_function


def get_metrics(metrics, unreplicate=False, stack=False):
    """
    The get_metrics function is a helper function that takes the metrics dictionary
    returned by the training loop and converts it to a format that can be used for
    plotting. It does this in two ways:

    :param metrics: Store the metrics that we want to track
    :param unreplicate: Convert the metrics from a replicated
    :param stack: Stack the metrics in a list
    :return: A dictionary of metrics
    
    """
    if unreplicate:
        metrics = flax.jax_utils.unreplicate(metrics)
    metrics = jax.device_get(metrics)
    if stack:
        return jax.tree_map(lambda *args: np.stack(args), *metrics)
    else:
        return {key: float(val) for key, val in metrics.items()}


def tree_path_to_string(path, sep=None):
    """
    The tree_path_to_string function takes a tree path and returns a string representation of it.

    :param path: Specify the path of the tree
    :param sep: Join the keys with a separator
    :return: A tuple of strings
    
    """
    keys = []
    for key in path:
        if isinstance(key, jax.tree_util.SequenceKey):
            keys.append(str(key.idx))
        elif isinstance(key, jax.tree_util.DictKey):
            keys.append(str(key.key))
        elif isinstance(key, jax.tree_util.GetAttrKey):
            keys.append(str(key.name))
        elif isinstance(key, jax.tree_util.FlattenedIndexKey):
            keys.append(str(key.key))
        else:
            keys.append(str(key))
    if sep is None:
        return tuple(keys)
    return sep.join(keys)


def flatten_tree(xs, is_leaf=None, sep=None):
    """
    The flatten_tree function takes a nested structure of arrays and returns a
    dictionary mapping from string keys to the corresponding array values. The
    string keys are derived from the tree path to each value, with `sep` used as
    the separator between levels in the tree. For example:

    :param xs: Store the tree structure
    :param is_leaf: Determine if a node is a leaf
    :param sep: Specify the separator between each key in the path
    :return: A dict of flattened tree paths to values
    
    """
    flattened, _ = jax.tree_util.tree_flatten_with_path(xs, is_leaf=is_leaf)
    output = {}
    for key, val in flattened:
        output[tree_path_to_string(key, sep=sep)] = val
    return output


def named_tree_map(f, tree, *rest, is_leaf=None, sep=None):
    """ An extended version of jax.tree_util.tree_map, where the mapped function
        f takes both the name (path) and the tree leaf as input.
    """
    return jax.tree_util.tree_map_with_path(
        lambda path, x, *r: f(tree_path_to_string(path, sep=sep), x, *r),
        tree, *rest,
        is_leaf=is_leaf
    )


def match_partition_rules(rules, params):
    """ Returns a pytree of PartitionSpec according to rules. Supports handling
        Flax TrainState and Optax optimizer state.
    """

    def get_partition_spec(name, leaf):
        if len(leaf.shape) == 0 or np.prod(leaf.shape) == 1:
            """ Don't partition scalar values. """
            return PS()
        for rule, ps in rules:
            if re.search(rule, name) is not None:
                return ps
        raise ValueError(f'Partition rule not found for param: {name}')

    return named_tree_map(get_partition_spec, params, sep='/')


def get_weight_decay_mask(exclusions):
    """ Return a weight decay mask function that computes the pytree masks
        according to the given exclusion rules.
    """

    def decay(name, _):
        for rule in exclusions:
            if re.search(rule, name) is not None:
                return False
        return True

    def weight_decay_mask(params):
        return named_tree_map(decay, params, sep='/')

    return weight_decay_mask


def tree_apply(fns, tree):
    """
    The tree_apply function is a generalization of the map function.
    It takes two arguments: a pytree of functions and a pytree of values.
    The tree_apply function applies each function in the first argument to its corresponding value in the second argument,
    and returns a new pytree with these results.

    :param fns: Apply the functions to the tree
    :param tree: Apply the function to each element in the tree
    :return: A pytree of the same structure as the input
    """
    return jax.tree_util.tree_map(lambda fn, x: fn(x), fns, tree)


def create_mesh(
        axis_dims: Sequence[int] = (1, -1, 1, 1), axis_names: Sequence[str] = ("dp", "fsdp", "tp", "sp"), backend=""
):
    """
    The create_mesh function creates a mesh object that can be used to shard arrays.

    :param axis_dims: Sequence[int]: Specify the dimensions of the mesh
    :param axis_names: Sequence[str]: Name the axes of the mesh
    :param backend: Specify the backend to use
    :return: A mesh object

    """
    array_devices = jax.numpy.ones((len(jax.devices() if backend == "" else jax.devices(backend)), 1))
    resh = array_devices.reshape(axis_dims).shape

    return jax.sharding.Mesh(
        create_device_mesh(resh), axis_names
    )



from setuptools import setup

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fjformer",
    version="0.0.21",
    author="Erfan Zare Chavoshi",
    author_email="erfanzare82@yahoo.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/erfanzar/fjformer",
    packages=setuptools.find_packages(),
    install_requires=[
        "chex",
        "typing",
        "jax>=0.4.17",
        "transformers>=4.34.0",
        "flax>=0.7.1",
        "optax>=0.1.7",
        "einops",
        "msgpack",
        "ml_collections",
        "einops",
        "datasets",
        "IPython"
    ],
    python_requires=">=3.7",
    license="Apache License 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



from .mesh_utils import (
    get_jax_mesh,
    names_in_current_mesh,
    get_names_from_partition_spec,
    match_partition_rules,
    flatten_tree,
    get_metrics,
    tree_apply,
    get_weight_decay_mask,
    named_tree_map,
    tree_path_to_string,
    make_shard_and_gather_fns,
    with_sharding_constraint,
    wrap_function_with_rng,
    create_mesh
)
